@@ -1,3 +1,6 @@
+# Copyright © 2009-2010 Pietro Battiston
+#   http://www.pietrobattiston.it
+#
 # Copyright (C) 2008 Brent Woodruff
 #   http://www.fprimex.com
 #
@@ -5,37 +8,52 @@
 #   http://garion.tzo.com/python/
 # 
 # This library is free software; you can redistribute it and/or
-# modify it under the terms of the GNU Lesser General Public
+# modify it under the terms of the GNU General Public
 # License as published by the Free Software Foundation; either
-# version 2.1 of the License, or (at your option) any later version.
+# version 2 of the License, or (at your option) any later version.
 # 
 # This library is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-# Lesser General Public License for more details.
+# General Public License for more details.
 # 
-# You should have received a copy of the GNU Lesser General Public
+# You should have received a copy of the GNU General Public
 # License along with this library; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-# 
-#
-# Change log:
-# 04-Aug-08 - Added Gallery2 compatibility
-#             Changed fetch_albums and fetch_albums_prune to return dicts
-#             Added docstrings
-#             Created package and registered with Pypi
-#
-# 09-Jun-04 - Removed self.cookie='' from _doRequest to allow multiple
-#             transactions for each login.
-#             Fixed cut paste error in newAlbum.
-#             (both patches from Yuti Takhteyev
 
-import urllib
+__version__ = '0.5'
+
+USER_AGENT = 'python-galleryremote v. %s' % (__version__)
+
+CA_SYSTEM_DIR = '/etc/ssl/certs'
+
 import urllib2
-import urllib2_file # REQUIRED for uploading pictures. Original source from:
-                    # http://fabien.seisen.org/python/
+
+try:
+    from M2Crypto import m2urllib2, SSL
+    M2CRYPTO_AVAILABLE = True
+except:
+    M2CRYPTO_AVAILABLE = False
+
 import StringIO
-import string
+import time
+from multipart import multipart
+
+class GalleryException(Exception):
+    """
+    The base class for exceptions related to error messages from gallery.
+    """
+
+class ConnectionException(Exception):
+    """
+    The base class for exceptions related to connection to gallery.
+    """
+
+class NotSupportedException(Exception):
+    """
+    For operations that only one version of Gallery supports, if asked with the
+    other version.
+    """
 
 class Gallery:
     """
@@ -63,6 +81,15 @@ class Gallery:
         url - base address of the gallery
         version - version of the gallery being connected to (default 2),
                   either 1 for Gallery1 or 2 for Gallery2
+        
+        gallery-uploader is able to cope with secured connections, thanks to
+        M2Crypto. It won't catch any exception, though: handling i.e. unverified
+        certificates is entirely left to higher level applications, which must
+        hence be ready to catch M2Crypto exceptions.
+        Moreover, gallery-uploader works fine (only emits a warning) also if
+        M2Crypto is not available, and in that case connections will NOT be
+        secure! If you want to avoid this risk, you can simply assert
+        gallery.M2CRYPTO_AVAILABLE .
         """
         self.version = version # Gallery1 or Gallery2
         if version == 1:
@@ -70,39 +97,85 @@ class Gallery:
         else:
             # default to G2
             self.url = url + '/main.php'
+        
+        # Until proven otherwise:
+        self.secured = False
+        
+        if self.url.startswith( 'https://' ):
+            if M2CRYPTO_AVAILABLE:
+                self.secured = True
+            else:
+                print "WARNING: M2Crypto missing, gallery-uploader connections will not be secure!"
+        
+        if self.secured:
+            self.ssl_context = SSL.Context()
+            self.ssl_context.load_verify_info( capath=CA_SYSTEM_DIR )
+            self.ssl_context.set_verify( SSL.verify_peer |
+                                         SSL.verify_fail_if_no_peer_cert |
+                                         SSL.verify_client_once, 20 )
+            self.opener = m2urllib2.build_opener( ssl_context=self.ssl_context )
+        else:
+            self.opener = urllib2.build_opener()
+
+            
         self.logged_in = 0
         self.cookie = ''
         self.protocol_version = '2.5'
+        self.auth_token = ''
 
-    def _do_request(self, request):
+    def _do_request(self, request, file_info=None):
         """
         Send a request, encoded as described in the Gallery Remote protocol.
         request - a dictionary of protocol parameters and values
+        file_info - a tuple with the field name and the filename to be added
+          in the body
         """
+        boundary = '------python-galleryremote_boundary_%d' % int(time.time())
+
+        headers = {'User-agent' : USER_AGENT,
+                   'Content-type':'multipart/form-data; boundary=%s' % boundary,
+                   'Accept': 'text/plain'}
+
+        if self.auth_token and self.version >= 2:
+            request['g2_authToken'] = self.auth_token
+
         if self.cookie != '':
-            headers = {'Cookie' : self.cookie}
-            enc_request = urllib.urlencode(request)
-            req = urllib2.Request(self.url, enc_request, headers)
-            response = urllib2.urlopen(req)
-        else:
-            enc_request = urllib.urlencode(request)
-            req = urllib2.Request(self.url, enc_request)
-            response = urllib2.urlopen(req)
+            headers['Cookie'] = self.cookie
+        enc_request = multipart(boundary, request, file_info)
+        req = urllib2.Request(self.url, enc_request, headers)
+        response = self.opener.open( req )
         
         info = response.info()
 
-        if info.has_key('set-cookie'):
-            self.cookie = info['set-cookie'].split(';')[0]
-
+        # There may be multiple "Set-cookie" fields, see:
+        # http://gallery.menalto.com/node/83512
+        # http://bugs.python.org/issue1660009
+        for header in info.getheaders('Set-cookie'):
+            # ... so we want to remember the last meaningful one:
+            sid_field = header.partition(';')[0]
+            if sid_field != 'GALLERYSID=':
+                self.cookie = sid_field
+        
         data = response.read()
         response = self._parse_response( data )
-        #print data
+        
+        if 'auth_token' in response:
+            self.auth_token = response['auth_token']
 
-        if response['status'] != '0':
-            raise response['status_text']
+        if 'status' in response:
+            # This _should be_ the right code to handle exceptions in Gallery 1,
+            # but I didn't test; it was already there and assumed that 'status'
+            # _was_ in response -- Pietro Battiston
+            if response['status'] != '0':
+                raise GalleryException, response['status_text']
+        else:
+            # ... but apparently, in Gallery 2 exceptions response _doesn't_
+            # contain 'status' (FIXME: VERIFY!):
+            if 'debug_exception' in response:
+                raise GalleryException, response['debug_exception']
 
         return response
-
+    
     def _parse_response(self, response):
         """
         Decode the response from a request, returning a request dict
@@ -111,25 +184,31 @@ class Gallery:
         """
         myStr = StringIO.StringIO(response)
         
+        header_found = False
+        
         for line in myStr:
-            if string.find( line, '#__GR2PROTO__' ) != -1:
+            if '#__GR2PROTO__' in line:
+                header_found = True
                 break
-                                            
-        # make sure the 1st line is #__GR2PROTO__
-        if string.find( line, '#__GR2PROTO__' ) == -1:
-            raise "Bad response: \r\n" + response
+        
+        if not header_found:
+            # the 1st line should start with #__GR2PROTO__ !
+            raise ConnectionException, "Bad response: \n" + response
             
         resDict = {}
         
+        # This operates still on myStr (on the lines _following_ the header):
         for myS in myStr:
-            strList = string.split(myS, '=', 2)
+            strList = myS.split('=', 2)
                 
             try:
-                resDict[strList[0]] = strList[1][:-1]
+                resDict[strList[0]] = strList[1]
+                if resDict[strList[0]].endswith('\n'):
+                    resDict[strList[0]] = resDict[strList[0]][:-1]                    
             except:
                 resDict[strList[0]] = ''
-                
-        return resDict         
+        
+        return resDict
 
     def _get(self, response, kwd):
         """
@@ -138,9 +217,9 @@ class Gallery:
             retval = response[kwd]
         except:
             retval = ''
-                        
-        return retval            
-
+        
+        return retval
+    
     def login(self, username, password):
         """
         Establish an authenticated session to the remote gallery.
@@ -163,10 +242,10 @@ class Gallery:
                 'g2_form[password]' : password
             }
         response = self._do_request(request)
-
+        
         # as long as it comes back here without an exception, we're ok.
-        self.logged_in = 1
-
+        self.logged_in = True
+    
     def fetch_albums(self):
         """
         Obtain a dict of albums contained in the gallery keyed by
@@ -202,11 +281,12 @@ class Gallery:
             album['perms.del_alb']          = self._get(response,'album.perms.del_alb.' + str(x))
             album['perms.create_sub']       = self._get(response,'album.perms.create_sub.' + str(x))
             album['perms.info.extrafields'] = self._get(response,'album.info.extrafields' + str(x))
+            album['ownerid']                = self._get(response,'album.ownerid' + str(x))
             
             albums[album['name']] = album
         
         return albums
-
+    
     def fetch_albums_prune(self):
         """
         Obtain a dict of albums contained in the gallery keyed by
@@ -248,6 +328,7 @@ class Gallery:
             album['perms.del_alb']          = self._get(response,'album.perms.del_alb.' + str(x))
             album['perms.create_sub']       = self._get(response,'album.perms.create_sub.' + str(x))
             album['perms.info.extrafields'] = self._get(response,'album.info.extrafields' + str(x))
+            album['ownerid']                = self._get(response,'album.ownerid' + str(x))
             
             albums[album['name']] = album
         
@@ -266,26 +347,24 @@ class Gallery:
                 'protocol_version' : self.protocol_version,
                 'cmd' : 'add-item',
                 'set_albumName' : album,
-                'userfile' : file,
                 'userfile_name' : filename,
                 'caption' : caption,
                 'extrafield.Description' : description
             }
+            response = self._do_request(request, ('userfile', filename) )
         else:
             request = {
                 'g2_controller' : 'remote:GalleryRemote',
                 'g2_form[protocol_version]' : self.protocol_version,
                 'g2_form[cmd]' : 'add-item',
                 'g2_form[set_albumName]' : album,
-                'g2_form[userfile]' : file,
                 'g2_form[userfile_name]' : filename,
                 'g2_form[caption]' : caption,
                 'g2_form[extrafield.Description]' : description
             }
-        file = open(filename)
-        response = self._do_request(request)
-        # if we get here, everything went ok.
+            response = self._do_request(request, ('g2_userfile', filename) )
         
+    
     def album_properties(self, album):
         """
         Obtain album property information for the specified album.
@@ -314,7 +393,33 @@ class Gallery:
             res_dict['add_to_beginning'] = response['add_to_beginning']
         
         return res_dict
-        
+
+    def image_properties(self, image):
+         """
+         Obtain image property information for the specified image.
+         image - the identifier of the image for which we're interested in
+         """
+         if self.version == 1:
+             raise NotSupportedException, "Operation not supported in Gallery version 1"
+         else:
+             request = {
+                 'g2_controller' : 'remote:GalleryRemote',
+                 'g2_form[protocol_version]' : self.protocol_version,
+                 'g2_form[cmd]' : 'image-properties',
+                 'g2_form[id]' : image
+             }
+         response = self._do_request(request)
+         
+         res_dict = {}
+         
+         for key in ['name', 'raw_height', 'raw_width', 'raw_filesize',\
+         'resizedName', 'resized_width', 'resized_height', 'thumbName',\
+         'thumb_width', 'thumb_height', 'caption', 'title', 'force', 'hidden']:
+             if response.has_key('image.'+key):
+                 res_dict[key] = response['image.'+key]
+                 
+         return res_dict
+    
     def new_album(self, parent, name=None, title=None, description=None):
         """
         Add an album to the specified parent album.
@@ -353,11 +458,11 @@ class Gallery:
         response = self._do_request(request)
         
         return response['album_name']
-        
+    
     def fetch_album_images(self, album):
         """
         Get the image information for all images in the specified album.
-        album - specifies the album from which to obtain image information 
+        album - specifies the album from which to obtain image information
         """
         if self.version == 1:
             request = {
@@ -404,8 +509,41 @@ class Gallery:
             image['capturedate.minutes'] = self._get(response, 'image.capturedate.minutes' + str(x))
             image['capturedate.seconds'] = self._get(response, 'image.capturedate.seconds' + str(x))
             image['description']         = self._get(response, 'image.extrafield.Description.' + str(x))
-            
+            image['hidden']              = self._get(response, 'image.hidden.' + str(x))
             images.append(image)
-            
-        return images 
         
+        return images
+    
+    def fetch_image(self, image, thumb=True):
+         """
+         Get the image (this returns the real image data).
+         If "thumb" is set, this will first do an "image-properties" request to
+         get the name of the thumbnail, then get the thumbnail (if a thumbnail is
+         not found, the image itself is returned).
+         Alternatively, the name of the thumbnail can be passed directly as
+         "image"; in this case, "thumb" must be False (since a thumbnail doesn't
+         have a thumbnail!).
+         
+         FIXME: fetch_album_images already gets the name of each thumbnail, and
+         it's very likely that it gets called before any call to fetch_image, so
+         some caching of the information could spare many requests.
+         """
+         if self.version == 1:
+             # I must search for the image url in Gallery1 - Pietro
+             raise NotImplementedError, "Image retrieval curretnly not implemented for Gallery 1"
+ 
+         if thumb:
+             image_info = self.image_properties(image)
+             if 'thumName' in image_info:
+                 image = image_info[thumbname]
+ 
+         image_url = self.url + '?g2_view=core.DownloadItem&g2_itemId=%s' % str(image)
+ 
+         headers = {}
+         if self.cookie != '':
+             headers['Cookie'] = self.cookie
+ 
+         req = urllib2.Request(image_url, headers=headers)
+         response = self.opener.open( req )
+ 
+         return response.read()
